#ifndef LIB_DBG_DEBUG_GLOBAL_PROGRAM_TRACER_MQH_INCLUDED
#define LIB_DBG_DEBUG_GLOBAL_PROGRAM_TRACER_MQH_INCLUDED
#property version "5.50"
/**********************************************************************************
 * Copyright (C) 2020 Dominik Egert <info@freie-netze.de>
 *
 * This file is the debugger main user macros include file.
 *
 * Lisence applied: GPLv2
 *                  https://www.gnu.org/licenses/old-licenses/gpl-2.0.html
 *
 * Author Dominik Egert / Freie Netze UG.
 **********************************************************************************
 *
 *  Version: 4120
 *  State: public
 *
 *  File information
 *  ================
 *
*/




//*********************************************************************************************************************************************************/
//  BEGIN Debugging macros
//


/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Dynamic heap allocation tracer
//

#ifdef LIB_DEBUG
    #ifdef LIB_DEBUG_TRACE_HEAP_MEMORY
        class __dbg_heap_obj_trace
        {
            public:
        
            // Local storage
            int     __allocated_bytes;
            int     __max_allocated_bytes;
            int     __max_simultaneous_objects;
            int     __objects_created;
            int     __objects_deleted;
            int     __file_ln;
            string  __file_name;
            string  __function;
            string  __obj_name;
            ulong   __sn_cnt;
            int     _size;
            int     _del_size;
            void*   v_obj[];
            ulong   v_obj_id[];
            int     v_obj_line[];
            string  v_obj_function[];
            string  v_obj_file[];
            string  v_obj_typename[];
            string  v_obj_obj_name[];
            void*   deleted_obj[];
            ulong   deleted_obj_id[];
            
            
            __dbg_heap_obj_trace() :
                __allocated_bytes           (NULL),
                __max_allocated_bytes       (NULL),
                __max_simultaneous_objects  (NULL),
                __objects_created           (NULL),
                __objects_deleted           (NULL),
                __sn_cnt                    (1),
                _size                       (NULL),
                _del_size                   (NULL)
            {};
        
            ~__dbg_heap_obj_trace()
            { 
                const int obj_count  =::ArraySize(v_obj_id);
                LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("### LIB_DEBUG ### Maximum assigned memory: %llu bytes", __max_allocated_bytes);
                LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("### LIB_DEBUG ### Maximum simultaneous existing objects: %llu", __max_simultaneous_objects);
                LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("### LIB_DEBUG ### Total objects created: %llu", __objects_created);
                LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("### LIB_DEBUG ### Total objects deleted: %llu", __objects_deleted);
                if(obj_count > NULL)
                {
                    string id_csv_list = NULL;
                    LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("%s", "### LIB_DEBUG ### Memory leakage report BEGIN");
                    LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("### LIB_DEBUG ### Currently undeleted objects: %i", ::ArraySize(v_obj_id));
                    for(int cnt = NULL; (cnt < obj_count) && !_StopFlag; cnt++)
                    { 
                        LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("    Object ID: %llu; created in file %s at line %i in function %s; Object type: %s", v_obj_id[cnt], v_obj_file[cnt], v_obj_line[cnt], v_obj_function[cnt], v_obj_typename[cnt]); 
                        id_csv_list += LIB_DBG_NAMESPACE(dbg_lib, dbg_StringFormat)("%llu, ", v_obj_id[cnt]);
                    }
                    LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("    Memory leakage object ID list: %s", StringSubstr(id_csv_list, 0, StringLen(id_csv_list) - 2));
                    LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("%s", "### LIB_DEBUG ### Memory leakage report END");
                    return;
                }
                LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("%s", "### LIB_DEBUG ### No memory leakage detected.");
            };
        
            __dbg_heap_obj_trace* _location_recorder(const int _line, const string _file, const string _function, string _obj_name = "NV")
            {
                __file_ln   = _line;
                __file_name = _file;
                __function  = _function;
                __obj_name  = _obj_name;
                return(::GetPointer(this));
            };
        
            template <typename V>
            const ulong identify(V* p_in)
            {
                int cnt = NULL;
                while( (cnt < _size)
                    && (v_obj[cnt] != p_in) )
                { cnt++; }
        
                if(cnt == _size)
                {
                    cnt = NULL;
                    while( (cnt < _del_size)
                        && (deleted_obj[cnt] != p_in) )
                    { cnt++; }
        
                    if(cnt < _del_size)
                    { return(deleted_obj_id[cnt]); }
                }
                else
                { return(v_obj_id[cnt]); }
        
                LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("Invalid object query: %s; Object not found!", typename(p_in));
                return(-1);
            };
        
            template <typename V>
            V* operator=(V* p_in)
            {
                _size = ::ArrayResize(v_obj, _size + 1);
                ::ArrayResize(v_obj_id, _size);
                ::ArrayResize(v_obj_line, _size);
                ::ArrayResize(v_obj_function, _size);
                ::ArrayResize(v_obj_file, _size);
                ::ArrayResize(v_obj_typename, _size);
                ::ArrayResize(v_obj_obj_name, _size);
                v_obj[_size - 1]            = p_in;
                v_obj_id[_size - 1]         = __sn_cnt++;
                v_obj_line[_size - 1]       = __file_ln;
                v_obj_function[_size - 1]   = __function;
                v_obj_file[_size - 1]       = __file_name;
                v_obj_typename[_size - 1]   = typename(V);
                v_obj_obj_name[_size - 1]   = __obj_name;
                __allocated_bytes          += sizeof(p_in);
                __objects_created++;
                __max_allocated_bytes       = (__max_allocated_bytes < __allocated_bytes) ? __allocated_bytes : __max_allocated_bytes;
                __max_simultaneous_objects  = (_size > __max_simultaneous_objects) ? _size : __max_simultaneous_objects;
                LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("%s", LIB_DBG_NAMESPACE(dbg_lib, dbg_context_msg)(DBG_OUTPUT_PREFIX, __file_name, __function, __file_ln, LIB_DBG_NAMESPACE(dbg_lib, dbg_StringFormat)("%s >>> new(%s) = 0x%016llX; Allocated Bytes: %i (Object ID: %llu)", "MQL5-API", typename(V), p_in, sizeof(p_in), v_obj_id[_size - 1]), "=", DBG_MSG_FORMAT_RIGHT_COLUMN_SPACER));
                
                return(p_in);
            }
        
            template <typename V>
            V* operator-(V* p_in)
            {
                int cnt = NULL;
                while( (cnt < _size)
                    && (v_obj[cnt] != p_in) )
                { cnt++; }
        
                if(cnt < _size)
                {
                    _del_size = ::ArrayResize(deleted_obj, _del_size + 1);
                    ::ArrayResize(deleted_obj_id, _del_size);
                    deleted_obj[_del_size - 1] = v_obj[cnt];
                    deleted_obj_id[_del_size - 1] = v_obj_id[cnt];    
                    LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("%s", LIB_DBG_NAMESPACE(dbg_lib, dbg_context_msg)(DBG_OUTPUT_PREFIX, __file_name, __function, __file_ln, LIB_DBG_NAMESPACE(dbg_lib, dbg_StringFormat)("%s >>> delete(%s) = 0x%016llX; Allocated Bytes: %i (Object ID: %llu)", DBG_MQLAPI_VERSION, typename(V), p_in, sizeof(p_in), v_obj_id[_size - 1]), "=", DBG_MSG_FORMAT_RIGHT_COLUMN_SPACER));
                    
                    ::ArrayRemove(v_obj, cnt, 1);
                    ::ArrayRemove(v_obj_id, cnt, 1);
                    ::ArrayRemove(v_obj_line, cnt, 1);
                    ::ArrayRemove(v_obj_function, cnt, 1);
                    ::ArrayRemove(v_obj_file, cnt, 1);
                    ::ArrayRemove(v_obj_typename, cnt, 1);
                    ::ArrayRemove(v_obj_obj_name, cnt, 1);
                    _size--;
                    __allocated_bytes          -= sizeof(p_in);
                    __objects_deleted++;

                    return(p_in);
                }
                LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("%s", LIB_DBG_NAMESPACE(dbg_lib, dbg_context_msg)(DBG_OUTPUT_PREFIX, __file_name, __function, __file_ln, LIB_DBG_NAMESPACE(dbg_lib, dbg_StringFormat)("%s >>> delete(%s) = %s; (Object ID not found)", DBG_MQLAPI_VERSION, typename(V), EnumToString(CheckPointer(p_in))), "=", DBG_MSG_FORMAT_RIGHT_COLUMN_SPACER));

                return(p_in);
            }
        } dbg_heap_obj_trace;

    #endif

#endif
//
/////////////////////////////////////////////////////////////////////////////////////////////////////




/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Global self destructing logging filehandle
//



#ifdef LIB_DBG_LOG_TO_FILE
    #ifndef __MQL4__
        namespace dbg_lib
        {

    #endif

        struct __dbg_file_handle
        {
            int obj;
            __dbg_file_handle() : obj(INVALID_HANDLE)   { obj = FileOpen(DBG_LOG_FILENAME, FILE_COMMON | FILE_TXT | FILE_ANSI | FILE_WRITE | FILE_SHARE_READ, CP_ACP); };
            ~__dbg_file_handle()                        { if(obj != INVALID_HANDLE) { FileClose(obj); } }
        } __dbg_f_h;

        void LIB_DBG_NAMESPACE_DEF(dbg_lib, dbg_print_to_file)(const string out)
        { FileWriteString(__dbg_f_h.obj, out + "\r\n", StringLen(out)); FileFlush(__dbg_f_h.obj); }

    #ifndef __MQL4__
        };

    #endif

#endif
//
/////////////////////////////////////////////////////////////////////////////////////////////////////




/////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Program lifetime tracer
//

#ifndef LIB_DEBUG_NAMESPACE
    #ifndef __MQL4__
        namespace dbg_lib
        {

    #endif

    const string LIB_DBG_NAMESPACE_DEF(dbg_lib, dbg_program_info)()
    {
        static bool call_once = false;

        if(call_once)
        { return(""); }
        call_once = true;
        return(DBG_STR_RESOLVE_EX45_FILEINFO);
    };

    #ifndef __MQL4__
        };

    #endif

#endif


#ifdef LIB_DEBUG
    struct __dbg_program_lifetime
    {
        ulong   program_start_time;
        string  program_init_name;

        __dbg_program_lifetime() :
            program_start_time(GetMicrosecondCount()),
            program_init_name(MQLInfoString(MQL_PROGRAM_NAME))
        { LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("### LIB_DEBUG ### \r\n%s\r\n### LIB_DEBUG ###", DBG_STR_EX45_FILEINFO); };

        ~__dbg_program_lifetime()
        {
            ulong micros = (GetMicrosecondCount() - program_start_time);
            string runtime = LIB_DBG_NAMESPACE(dbg_lib, dbg_StringFormat)("Minutes: %llu; Seconds: %i; Microseconds: %llu", micros / 60000000, (micros / 1000000) % 60, micros % 1000000);
            string uninit_txt = LIB_DBG_NAMESPACE(dbg_lib, uninit_text)("@GlobalSpace");
            uninit_txt = (uninit_txt == NULL) && MQLInfoInteger(MQL_DEBUG) ? "Debugger stopped!" : uninit_txt;
            uninit_txt = (uninit_txt == NULL) && (_UninitReason == 0) ? "ExpertRemove() called!" : uninit_txt;
            uninit_txt = (uninit_txt == NULL) ? "Abnormal termination!" : uninit_txt;
            #ifndef __MQL4__
                LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("\n### LIB_DEBUG ### Program @GlobalSpace lifetime end. Runtime: %s.\r\n### LIB_DEBUG ### %s\r\n### LIB_DEBUG ### %s", runtime, uninit_txt, program_init_name);

            #else
                LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("### LIB_DEBUG ### %s", program_init_name); 
                LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("### LIB_DEBUG ### %s", uninit_txt);
                LIB_DBG_NAMESPACE(dbg_lib, dbg_printf)("### LIB_DEBUG ### Program @GlobalSpace lifetime end. Runtime: %s.", runtime);

            #endif
        };
    } __dbg_program_session_lifetime();

#endif
//
/////////////////////////////////////////////////////////////////////////////////////////////////////


//
// END Debugging support
//*********************************************************************************************************************************************************/
#endif // LIB_DBG_DEBUG_GLOBAL_PROGRAM_TRACER_MQH_INCLUDED
